searchData={"content_type":"text/plain","items":[{"doc":null,"ref":"nine.html","title":"nine","type":"module"},{"doc":null,"ref":"nine.html#compile/1","title":"nine.compile/1","type":"function"},{"doc":null,"ref":"nine.html#compile/3","title":"nine.compile/3","type":"function"},{"doc":"# nine\n\nA library providing a data driven router compiler. \n\nThe goal of `nine` is to allow developers to compose middleware and handlers.\nThis should significantly decrease boiler plate and give a web development experience competitive with other\nlanguages and ecosystems.","ref":"readme.html","title":"nine","type":"extras"},{"doc":"rebar3 compile","ref":"readme.html#build","title":"Build - nine","type":"extras"},{"doc":"`nine` functions as a router compiler library. Meaning you give it a router config and it compiles a router module.\n`nine` does not have it's own web server, it is meant to be paired with a backend.\n\n`nine` provides one single function to be used: `compile`. Compile can be given one map argument or 3 arguments.\n\n    nine:compile(#{routes => Routes, router => Router, generator => Generator})\n\nOR\n\n    nine:compile(Routes, Router, Generator).\n\nThe generator is provided by the backend, and is a module required to have at least one function `generate/2`.\nThe name of the `Router` module and the normalized routes after nine compiles the routes config given.\n\n`nine:compile` takes a router config and compiles it into an Erlang module at runtime using [forms](https://www.erlang.org/doc/apps/erts/absform).","ref":"readme.html#usage","title":"Usage - nine","type":"extras"},{"doc":"A route is a map with these keys:\n\n    [{ > => ...,\n       > => ...,\n       > => ...,\n       > => ...,\n       > => ...}]\n\npath: The URL path for the request to be handled. Example:  >.\nmethod: The request method to be handled. Example:  >.\npre: The middleware to be called before the request handler. Example: [{todo_mid, json_request}]\npost: The middleware to be called after the request handler. Example: [{todo_mid, json_response}]\nhandle: The request handler function.\n\nA minimum config might look like this:\n\n    [#{ > =>  >,\n        > => {index_handler, index}}]\n\nA full config for a todo application would look like this:\n\n    [#{ > =>  > =>\n        > =>  >,\n        > => {todo_handler, index}},\n     #{ > =>  >,\n        > => [{todo_mid, json_response}],\n        > => [\n         #{ > =>  >,\n            > => [{todo_mid, json_request}],\n            > => [\n             #{ > =>  >,\n                > => {todo_handler, post_todo}},\n             #{ > =>  >,\n                > => {todo_handler, delete_todo}}\n           ],\n         #{ > =>  >,\n            > =>  >,\n            > => {todo_handler, get_todos}}\n       ]\n     },\n     #{ > =>  >, \n        > => {todo_handler, not_found}}\n    ]\n\nThere is a lot going on here with this config which we will break down in the next sections.\n\n#","ref":"readme.html#routes","title":"Routes - nine","type":"extras"},{"doc":"A handler is specified as {module, function}. Example:\n\n    {todo_handler, get_todos}\n\n`todo_handler` being the module, and `get_todos` is the function.\n\nThe handler function is expected to look like this:\n\n    get_todos(Context) ->\n        Context#{resp => thoas:encode([#{id => 1, body =>  >}])}.\n\n##","ref":"readme.html#handler","title":"Handler - nine","type":"extras"},{"doc":"It is possible to compose complex router configs by nesting handlers. The `handle` key can either take a \ntuple or a list of route configs.\n\nFor example an `api` handler can branch off with multiple other request paths.\n\n    [{ > =>  >,\n       > => [\n        #{ > =>  >,\n           > =>  >,\n           > => {todo_handler, post_todo}},\n        #{ > =>  >,\n           > =>  >,\n           > => {todo_handler, get_todos}}\n    ]}]\n\nThis means when a POST request goes to `/api/todo` the function `todo_handler:post_todo/1` will be called. While\na request going to `/api/todos` will trigger the function `todo_handler:get_todos/1`.\n\n##","ref":"readme.html#nested-handlers","title":"Nested Handlers - nine","type":"extras"},{"doc":"`nine` builds in a way to have named parameters in the URL.\n\nA path like `/todo/:id` will result in the context map including the params key.\nThe value of the params will be `#{id =>  >}`.\nIn case you are worried about atoms coming from user data, it is okay for `id` to be an atom because it is a static value set at compile time. \n\nnine also supports partial path params like `/person/num:ber` will result in a params map looking like `#{ber =>  >}` for example. This is similar to how Phoenix works with routing.\n\n##","ref":"readme.html#url-path-params","title":"URL Path Params - nine","type":"extras"},{"doc":"`nine` supports catch all routes with `*` in the path. For example: ` >` will match any route.\n\nWe can also put a wildcard at the suffix of a path: ` >` will match a route like ` >`.\nWildcards can only be at the end of a path. This is similar to how Phoenix works with catch all routes.\nThe reason for this is the routing uses Erlang pattern matching, so it must follow the same rules.\n\n#","ref":"readme.html#wildcard","title":"Wildcard - nine","type":"extras"},{"doc":"Middleware are specified just like handlers, in fact they are the same thing! An example middleware might look like:\n\n    {nine_mid, json_response}\n\nMiddleware are functions that take a Context as input and output a Context or an elli response. One could write a logging middleware like this:\n\n    logging_middleware(Context) ->\n        logger:debug(#{context => Context}),\n        Context.\n\nOr we could make a middleware that adds some data to the Context:\n\n    message_middleware(Context) ->\n        Context#{message =>  >}.\n\nMiddleware are helpful in all sorts of situations and allow developers to write web apps in a DRY way.\n\n##","ref":"readme.html#middleware","title":"Middleware - nine","type":"extras"},{"doc":"`nine` specifies middleware chaining with the `pre` and `post` keys. Nesting routes will also concatenate the `pre` and `post` keys\nin the expected order.\n\nFor example:\n\n    #{ > => [{todo_mid, json_request}],\n       > => {todo_handler, post_todo}}\n\nWill generate a sequence of function calls where `todo_mid:json_request` is called first, and the result is passed to\n`todo_handler:post_todo`.\n\nAllowing `post_todo` to be implemented like so:\n\n    post_todo(Context=#{json := #{ > := Body}}) ->\n        todo_db:insert(Body),\n        nine_util:redirect(Context,  >).\n\n`post_todo` can expect the `json` key to be filled with data because `json_request` is called first.\n\n##","ref":"readme.html#middleware-chains","title":"Middleware Chains - nine","type":"extras"},{"doc":"There are situations where we want to return a response immediately without finishing the middleware chain. This is known as halting.\n\n`nine` makes this possible because each middleware and handler call is wrapped in a case statement checking for the `resp` key.\n\nIf a handler or middleware returns a Context map with the `resp` key it will immediately be sent without triggering further middleware.","ref":"readme.html#halting","title":"Halting - nine","type":"extras"},{"doc":"`nine` was inspired by other composable middleware tools.\n\n- [ring](https://github.com/ring-clojure/ring/wiki/Concepts) - Standard Clojure HTTP abstraction for web servers\n- [ataraxy](https://github.com/weavejester/ataraxy) - data driven routing library for Clojure\n- [Plug.Router](https://hexdocs.pm/plug/readme.html#plug-router) - Ecosystem defining Elixir HTTP middleware\n- [golang http middleware](https://dev.to/theghostmac/understanding-and-crafting-http-middlewares-in-go-3183) - Standard Library Golang Middleware Pattern\n- [Cowboy Router](https://ninenines.eu/docs/en/cowboy/2.10/guide/routing/) - cowboy router is compiled into a lookup table","ref":"readme.html#inspirations","title":"Inspirations - nine","type":"extras"},{"doc":"- The name `nine` comes from \"nine nines\".\n- Middleware was originally intended to look like Ring's, but wasn't compatible with Erlang's pattern matching lookups.","ref":"readme.html#fun-facts","title":"Fun Facts - nine","type":"extras"},{"doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS\n\n   Copyright 2023, Harley Swick  .\n\n   Licensed under the Apache License, Version 2.0 (the \"License\");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an \"AS IS\" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.","ref":"license.html","title":"LICENSE","type":"extras"}]}